import json
import datetime
import os
from typing import List

import pandas as pd

from config import TEMPORARY_OUTPUT_PATH
from study1.s1_config import TRUE_STRING, FALSE_STRING, USE_ABS_ERROR, \
    QUESTION_TYPE_LOOKUP_COMBINED, IGNORE_FILE_NR, T_TYPES, VIS_STRING, QUESTION_STRING, INPUT_PATH
from study1.base_data_processing import calc_correct_answers, get_time_data, calc_answers
from study1.s1_utils import get_question_type, get_block_order, convert_order_values
from utils import find_indices, data_string_converter


def get_errors(filter_methods: List = None, reload_csv: bool = False):
    filter_methods = filter_methods or []

    file_path = f"{TEMPORARY_OUTPUT_PATH}/S1-calcualted-data.csv"
    if os.path.isfile(file_path) and not reload_csv:
        error_df = pd.read_csv(file_path)
        error_df["calculatedValue"] = error_df.apply(lambda row: data_string_converter(row["calculatedValue"]), axis=1)
        error_df["trimmedValue"] = error_df.apply(lambda row: data_string_converter(row["trimmedValue"]), axis=1)
    else:
        output = []

        correct_answers_df = calc_correct_answers()
        user_answers_df = calc_answers()
        block_order = get_block_order()

        for i, row in user_answers_df.iterrows():
            if row.fileNr in IGNORE_FILE_NR:
                continue

            correct_answer = correct_answers_df[(correct_answers_df.block == row.block) & (correct_answers_df.questionNr == row.questionNr) & (correct_answers_df.note == row.note)].value.values[0]
            if str is type(correct_answer) and ("[" in correct_answer or "{" in correct_answer):
                try:
                    correct_answer = json.loads(correct_answer)
                except:
                    pass

            user_answer = row.value
            if user_answer in ["n.v.", "-"]:
                continue
            elif (row.blockType, row.questionNr) in QUESTION_TYPE_LOOKUP_COMBINED["Extremum"]:
                result = TRUE_STRING if user_answer in correct_answer else FALSE_STRING
            elif (row.blockType, row.questionNr) in QUESTION_TYPE_LOOKUP_COMBINED["Find Anomalies"] + QUESTION_TYPE_LOOKUP_COMBINED["Filter"]:
                result = TRUE_STRING if user_answer == correct_answer else FALSE_STRING
            elif row.blockType is 4 and row.questionNr is 4:
                try:
                    last_answer = user_answers_df[(user_answers_df.block == row.block) & (user_answers_df.questionNr == 2) & (user_answers_df.fileNr == row.fileNr)].value.values[0]
                    result = float(correct_answer[str(int(last_answer) - 1)]) - user_answer
                    result = abs(result) if USE_ABS_ERROR else result
                except KeyError:
                    result = None
            # Following this line: Only check for the types of both, correct and given answer.
            elif type(correct_answer) in [int, float] and type(user_answer) in [int, float]:
                result = correct_answer - user_answer
                result = abs(result) if USE_ABS_ERROR else result
            elif type(correct_answer) is list and type(user_answer) in [int, float]:
                result = TRUE_STRING if user_answer in correct_answer else FALSE_STRING
            elif type(correct_answer) is str and type(user_answer) is str:
                result = TRUE_STRING if correct_answer.strip() == user_answer.strip() else FALSE_STRING
            else:
                result = TRUE_STRING if user_answer == correct_answer else FALSE_STRING

            output.append((row, result))

        output_dict = {
            "order": [block_order[r.fileNr][r.block] for (r, e) in output],
            "blockType": [r.blockType for (r, e) in output],
            "block": [r.block for (r, e) in output],
            "complexity": [r.complexity for (r, e) in output],
            "questionNr": [r.questionNr for (r, e) in output],
            "questionKey": [f"{r.blockType}-{r.questionNr}" for (r, e) in output],
            "questionType": [get_question_type(r.blockType, r.questionNr, r.note, QUESTION_TYPE_LOOKUP_COMBINED) for (r, e) in output],
            "note": [r.note for (r, e) in output],
            "fileNr": [r.fileNr for (r, e) in output],
            "background": [r.background for (r, e) in output],
            "valueType": ["NUMERIC" if type(e) is not str else "BOOL" for (r, e) in output],
            "calculatedValue": [e for (r, e) in output],
        }

        error_df = pd.DataFrame(output_dict)
        error_df = convert_order_values(error_df)
        error_df = error_df.replace(TRUE_STRING, 1)
        error_df = error_df.replace(FALSE_STRING, 0)

        error_mean = error_df.groupby(by=["questionType"]).mean().reset_index()
        error_std = error_df.groupby(by=["questionType"]).std().reset_index()
        trimmed_values = []
        for i, row in error_df.iterrows():
            if type(row["calculatedValue"]) not in [float, int]:
                trimmed_values.append(row["calculatedValue"])
            new_value = float(error_mean[error_mean.questionType == row["questionType"]]["calculatedValue"] +
                              2 * error_std[error_std.questionType == row["questionType"]]["calculatedValue"])
            trimmed_values.append(row["calculatedValue"] if row["calculatedValue"] <= new_value else new_value)

        error_df["trimmedValue"] = trimmed_values
        error_df.to_csv(file_path)

    for method in filter_methods:
        error_df = method(error_df)

    return error_df


def get_times(filter_methods: List = None, reload_csv: bool = False) -> pd.DataFrame:
    filter_methods = filter_methods or []

    os.makedirs(TEMPORARY_OUTPUT_PATH, exist_ok=True)
    file_path = f"{TEMPORARY_OUTPUT_PATH}/S1-derived-times-data.csv"
    if os.path.isfile(file_path) and not reload_csv:
        times_df = pd.read_csv(file_path)
        times_df["calculatedValue"] = times_df.apply(lambda row: data_string_converter(row["calculatedValue"]), axis=1)
        times_df["trimmedValue"] = times_df.apply(lambda row: data_string_converter(row["trimmedValue"]), axis=1)
    else:
        time_data = get_time_data()
        block_order = get_block_order()

        output = {}

        for (file_nr, background_name, block_type, block, question_nr), rows in time_data.items():
            if file_nr in IGNORE_FILE_NR:
                continue
            if len(rows) <= 1:
                for t_type in T_TYPES:
                    output[(file_nr, background_name, block_type, block, question_nr, t_type)] = None
                continue

            t_start = datetime.datetime.strptime(f"{rows[0][0]} {rows[0][1]}", "%Y-%m-%d %H:%M:%S.%f")
            t_first_toggle = datetime.datetime.strptime(f"{rows[1][0]} {rows[1][1]}", "%Y-%m-%d %H:%M:%S.%f")
            answer_indices = find_indices(rows, "Answering", 3)
            t_answer = datetime.datetime.strptime(f"{rows[answer_indices[0]][0]} {rows[answer_indices[0]][1]}", "%Y-%m-%d %H:%M:%S.%f")
            t_finished_answer = datetime.datetime.strptime(f"{rows[-1][0]} {rows[-1][1]}", "%Y-%m-%d %H:%M:%S.%f")

            toggle_counter = -1
            for i in range(0, answer_indices[0]):
                if "Toggle" in rows[i][3]:
                    toggle_counter += 1
            toggle_counter_answering = 0
            for i in range(answer_indices[0], len(rows)):
                if "Toggle" in rows[i][3]:
                    toggle_counter_answering += 1
            endstate = VIS_STRING if toggle_counter % 2 == 0 else QUESTION_STRING
            state_dur = {0: datetime.timedelta(0), 1: datetime.timedelta(0)}  # 0 -> vis, 1 -> question
            toggle_state, last_index = 0, 1
            for i in range(2, answer_indices[0] + 1):
                t0 = datetime.datetime.strptime(f"{rows[last_index][0]} {rows[last_index][1]}", "%Y-%m-%d %H:%M:%S.%f")
                t1 = datetime.datetime.strptime(f"{rows[i][0]} {rows[i][1]}", "%Y-%m-%d %H:%M:%S.%f")
                state_dur[toggle_state] += t1 - t0
                if "Toggle" in rows[i][3]:
                    toggle_state = (toggle_state + 1) % 2
                last_index = i

            state_answer_dur = {0: datetime.timedelta(0), 1: datetime.timedelta(0)}  # 0 -> vis, 1 -> question
            toggle_state, last_index = toggle_state, answer_indices[0]
            for i in range(answer_indices[0] + 1, len(rows)):
                t0 = datetime.datetime.strptime(f"{rows[last_index][0]} {rows[last_index][1]}", "%Y-%m-%d %H:%M:%S.%f")
                t1 = datetime.datetime.strptime(f"{rows[i][0]} {rows[i][1]}", "%Y-%m-%d %H:%M:%S.%f")
                state_answer_dur[toggle_state] += t1 - t0
                if "Toggle" in rows[i][3]:
                    toggle_state = (toggle_state + 1) % 2
                last_index = i

            results = {
                T_TYPES[0]: t_first_toggle - t_start,
                T_TYPES[1]: t_answer - t_first_toggle,
                T_TYPES[2]: toggle_counter,
                T_TYPES[3]: state_dur[0],
                T_TYPES[4]: state_dur[1],
                T_TYPES[5]: endstate,
                T_TYPES[6]: t_finished_answer - t_answer,
                T_TYPES[7]: toggle_counter_answering,
                T_TYPES[8]: state_answer_dur[0],
                T_TYPES[9]: state_answer_dur[1],
                T_TYPES[10]: VIS_STRING if (toggle_counter + toggle_counter_answering) % 2 == 0 else QUESTION_STRING
            }

            for t_type in T_TYPES:
                output[(file_nr, background_name, block_type, block, question_nr, t_type)] = results[t_type]

        output_dict = {
            "order": [block_order[k[0]][k[3]] for k, v in output.items()],
            "blockType": [k[2] for k, v in output.items()],
            "block": [k[3] for k, v in output.items()],
            "complexity": [int(k[3] / 5 + 1) for k, v in output.items()],
            "questionNr": [k[4] for k, v in output.items()],
            "questionKey": [f"{k[2]}-{k[4]}" for k, v in output.items()],
            "questionType": [get_question_type(k[2], k[4], k[5], QUESTION_TYPE_LOOKUP_COMBINED) for k, v in output.items()],
            "note": [k[5] for k, v in output.items()],
            "fileNr": [k[0] for k, v in output.items()],
            "background": [k[1] for k, v in output.items()],
            "valueType": ["BOOL" if type(v) is str else ("TOGGLE" if "Toggle" in k[5] else "TIME")
                          for k, v in output.items()],
            "calculatedValue": [v if type(v) is not datetime.timedelta else v.total_seconds()
                      for k, v in output.items()],
        }

        times_df = pd.DataFrame(output_dict)
        times_df = convert_order_values(times_df)
        times_df = times_df.replace(VIS_STRING, 1)
        times_df = times_df.replace(QUESTION_STRING, 0)

        times_mean = times_df.groupby(by=["questionType", "note"]).mean().reset_index()
        times_std = times_df.groupby(by=["questionType", "note"]).std().reset_index()
        trimmed_values = []
        for i, row in times_df.iterrows():
            if type(row["calculatedValue"]) not in [float, int]:
                trimmed_values.append(row["calculatedValue"])
            new_value = float(times_mean[(times_mean.questionType == row["questionType"]) & (times_mean.note == row["note"])]["calculatedValue"] +
                              2 * times_std[(times_std.questionType == row["questionType"]) & (times_std.note == row["note"])]["calculatedValue"])
            trimmed_values.append(row["calculatedValue"] if row["calculatedValue"] <= new_value else new_value)

        times_df["trimmedValue"] = trimmed_values
        times_df.to_csv(file_path)

    for method in filter_methods:
        times_df = method(times_df)

    return times_df


e = get_errors()
t = get_times()
