import datetime
import os
from collections import Counter, OrderedDict
from typing import Dict, Tuple, List

import math
import matplotlib.pyplot as plt
import pandas as pd
import pingouin as pg
import seaborn as sns
from matplotlib.gridspec import GridSpec

from config import VALUE_COLUMN, VIS_FORMAT, FIG_SIZE, DPI, TOP_OFFSET, BOTTOM_OFFSET, SNS_AXES_STYLE, \
    PERCEPTION_RATING_COLOR_PALETTE, H_SPACE, BG_COLORS, FOCUS_COLORS, BG_F_COLORS
from study2.base_data_processing import get_questionnaire_data
from study2.data_processing import calc_values, calc_time_values
from study2.s2_config import VisConfigValues, IMAGE_OUTPUT_PATH, QUESTION_MAPPING, FOCUS_QUESTION_MAPPING
from utils import add_xaxis_labels, add_label_to_bars, set_legend_from_vis, set_rc_plot_values, set_custom_legend, \
    add_yaxis_labels_hue, add_caption


def background_questionType(error_df: pd.DataFrame, times_df: pd.DataFrame):
    hue_order = ["BG3", "BG4"]

    numeric_df = error_df[error_df.valueType == "NUMERIC"]
    bool_df = error_df[error_df.valueType == "BOOL"]
    bool_df[VALUE_COLUMN] = bool_df.apply(lambda row: 1 - row[VALUE_COLUMN], axis=1)

    fig = plt.figure(figsize=(FIG_SIZE[0], FIG_SIZE[1] * 1.1), dpi=DPI)
    gs = GridSpec(1, 19, figure=fig, top=TOP_OFFSET, bottom=BOTTOM_OFFSET / 1.1, left=0.02, right=0.99, wspace=1.5)

    sGs = gs.new_subplotspec(loc=[0, 0], rowspan=1, colspan=4).subgridspec(nrows=1, ncols=4, wspace=0.025)
    # Numeric View - Total
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 0])
        sns.barplot(y=VALUE_COLUMN, x="valueType", hue="background", data=numeric_df, ax=ax,
                    palette={k: BG_COLORS[k] for k in hue_order},
                    hue_order=hue_order, capsize=0.075, errwidth=VisConfigValues.ERROR_WIDTH, errcolor=VisConfigValues.ERROR_COLOR)

    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("Absolute Errors")
    ax.set_xticklabels(VisConfigValues.VIS_NUMERIC_ORDER)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VisConfigValues.VIS_VALUES["EN-ylim"])
    ax.set_yticks(VisConfigValues.VIS_VALUES["EN-yTicks"])
    ax.set_yticklabels(VisConfigValues.VIS_VALUES["EN-yLabels"])
    add_xaxis_labels(ax, labels=["Total"], pos="bottom")

    # Numeric View - Analysis Task
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 1:])
        sns.barplot(y=VALUE_COLUMN, x="questionType", hue="background", data=numeric_df, ax=ax,
                    palette={k: BG_COLORS[k] for k in hue_order},
                    hue_order=hue_order, order=VisConfigValues.VIS_NUMERIC_ORDER, capsize=0.075, errwidth=VisConfigValues.ERROR_WIDTH, errcolor=VisConfigValues.ERROR_COLOR)

    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("")
    ax.set_xticklabels(VisConfigValues.VIS_NUMERIC_ORDER)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VisConfigValues.VIS_VALUES["EN-ylim"])
    ax.set_yticks(VisConfigValues.VIS_VALUES["EN-yTicks"])
    ax.set_yticklabels([""] * len(VisConfigValues.VIS_VALUES["EN-yLabels"]))
    add_xaxis_labels(ax, labels=VisConfigValues.VIS_NUMERIC_ORDER, pos="bottom")

    sGs = gs.new_subplotspec(loc=[0, 4], rowspan=1, colspan=6).subgridspec(nrows=1, ncols=6, wspace=0.025)
    # Bool View - Total
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 0])
        sns.barplot(y=VALUE_COLUMN, x="valueType", hue="background", data=bool_df, ax=ax,
                    palette={k: BG_COLORS[k] for k in hue_order},
                    hue_order=hue_order, capsize=0.075, errwidth=VisConfigValues.ERROR_WIDTH, errcolor=VisConfigValues.ERROR_COLOR)

    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("Percentage Error")
    ax.set_xticklabels(VisConfigValues.VIS_BOOL_ORDER)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VisConfigValues.VIS_VALUES["EB-ylim"])
    ax.set_yticks(VisConfigValues.VIS_VALUES["EB-yTicks"])
    ax.set_yticklabels(VisConfigValues.VIS_VALUES["EB-yLabels"])
    add_xaxis_labels(ax, labels=["Total"], pos="bottom")

    # Bool View - Analysis Task
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 1:])
        sns.barplot(y=VALUE_COLUMN, x="questionType", hue="background", data=bool_df, ax=ax,
                    palette={k: BG_COLORS[k] for k in hue_order},
                    hue_order=hue_order, order=VisConfigValues.VIS_BOOL_ORDER, capsize=0.075, errwidth=VisConfigValues.ERROR_WIDTH, errcolor=VisConfigValues.ERROR_COLOR)

    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("")
    ax.set_xticklabels(VisConfigValues.VIS_BOOL_ORDER)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VisConfigValues.VIS_VALUES["EB-ylim"])
    ax.set_yticks(VisConfigValues.VIS_VALUES["EB-yTicks"])
    ax.set_yticklabels([""] * len(VisConfigValues.VIS_VALUES["EB-yLabels"]))
    add_xaxis_labels(ax, labels=VisConfigValues.VIS_BOOL_ORDER, pos="bottom")

    times_df = times_df[times_df.valueDescription == "SolvingDuration"]

    sGs = gs.new_subplotspec(loc=[0, 10], rowspan=1, colspan=9).subgridspec(nrows=1, ncols=9, wspace=0.025)
    # Time - Total
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 0])
        sns.barplot(y=VALUE_COLUMN, x="valueType", hue="background", data=times_df, ax=ax,
                    palette={k: BG_COLORS[k] for k in hue_order},
                    hue_order=hue_order, capsize=0.075, errwidth=VisConfigValues.ERROR_WIDTH, errcolor=VisConfigValues.ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("Completion Time in s")
    ax.set_xticklabels(VisConfigValues.QUESTION_TYPES)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VisConfigValues.VIS_VALUES["T-ylim"])
    ax.set_yticks(VisConfigValues.VIS_VALUES["T-yTicks"])
    ax.set_yticklabels(VisConfigValues.VIS_VALUES["T-yLabels"])
    add_xaxis_labels(ax, labels=["Total"], pos="bottom")

    # Add significant asterisks to the vis
    for i, (source, target, text) in enumerate([(1, 2, "*")]):
        b1 = ax.containers[source - 1][0]
        b2 = ax.containers[target - 1][0]
        height = ax.get_ylim()[1]
        x1 = b1.get_x() + b1.get_width() / 2.
        x2 = b2.get_x() + b2.get_width() / 2.
        y = b1.get_y() + height * 0.05 + height * 0.05 * i
        ax.plot([x1, x2], [y, y], linewidth=2, c="black")
        ax.text(x1 + (x2 - x1) / 2., y + 0.2, text, ha="center", va="center", color="black", weight="bold")

    # Time - Analysis Task
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 1:])
        sns.barplot(y=VALUE_COLUMN, x="questionType", hue="background", data=times_df, ax=ax,
                    palette={k: BG_COLORS[k] for k in hue_order},
                    hue_order=hue_order, order=VisConfigValues.QUESTION_TYPES, capsize=0.075, errwidth=VisConfigValues.ERROR_WIDTH, errcolor=VisConfigValues.ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("")
    ax.set_xticklabels(VisConfigValues.QUESTION_TYPES)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VisConfigValues.VIS_VALUES["T-ylim"])
    ax.set_yticks(VisConfigValues.VIS_VALUES["T-yTicks"])
    ax.set_yticklabels([""] * len(VisConfigValues.VIS_VALUES["T-yLabels"]))
    add_xaxis_labels(ax, labels=VisConfigValues.QUESTION_TYPES, pos="bottom")

    add_caption(fig, "A")
    set_legend_from_vis(fig, ax)
    fig.tight_layout()
    for format_str in VIS_FORMAT:
        fig.savefig(f"{IMAGE_OUTPUT_PATH}/background-questionType.{format_str}")
    plt.close(fig)


def focus_questionType(error_df: pd.DataFrame, times_df: pd.DataFrame):
    hue_order = ["Single Focus", "Split Focus"]

    numeric_df = error_df[error_df.valueType == "NUMERIC"]
    bool_df = error_df[error_df.valueType == "BOOL"]
    bool_df[VALUE_COLUMN] = bool_df.apply(lambda row: 1 - row[VALUE_COLUMN], axis=1)

    fig = plt.figure(figsize=(FIG_SIZE[0], FIG_SIZE[1] * 1.1), dpi=DPI)
    gs = GridSpec(1, 19, figure=fig, top=TOP_OFFSET, bottom=BOTTOM_OFFSET / 1.1, left=0.02, right=0.99, wspace=1.5)

    sGs = gs.new_subplotspec(loc=[0, 0], rowspan=1, colspan=4).subgridspec(nrows=1, ncols=4, wspace=0.025)
    # Numeric View - Total
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 0])
        sns.barplot(y=VALUE_COLUMN, x="valueType", hue="focus", data=numeric_df, ax=ax,
                    palette={k: FOCUS_COLORS[k] for k in hue_order},
                    hue_order=hue_order,
                    capsize=0.075, errwidth=VisConfigValues.ERROR_WIDTH, errcolor=VisConfigValues.ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("Absolute Errors")
    ax.set_xticklabels(VisConfigValues.VIS_NUMERIC_ORDER)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VisConfigValues.VIS_VALUES["EN-ylim"])
    ax.set_yticks(VisConfigValues.VIS_VALUES["EN-yTicks"])
    ax.set_yticklabels(VisConfigValues.VIS_VALUES["EN-yLabels"])
    add_xaxis_labels(ax, labels=["Total"], pos="bottom")

    # Numeric View - Analysis Task
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 1:])
        sns.barplot(y=VALUE_COLUMN, x="questionType", hue="focus", data=numeric_df, ax=ax,
                    palette={k: FOCUS_COLORS[k] for k in hue_order},
                    hue_order=hue_order, order=VisConfigValues.VIS_NUMERIC_ORDER,
                    capsize=0.075, errwidth=VisConfigValues.ERROR_WIDTH, errcolor=VisConfigValues.ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("")
    ax.set_xticklabels(VisConfigValues.VIS_NUMERIC_ORDER)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VisConfigValues.VIS_VALUES["EN-ylim"])
    ax.set_yticks(VisConfigValues.VIS_VALUES["EN-yTicks"])
    ax.set_yticklabels([""] * len(VisConfigValues.VIS_VALUES["EN-yLabels"]))
    add_xaxis_labels(ax, labels=VisConfigValues.VIS_NUMERIC_ORDER, pos="bottom")

    sGs = gs.new_subplotspec(loc=[0, 4], rowspan=1, colspan=6).subgridspec(nrows=1, ncols=6, wspace=0.025)
    # Bool View - Total
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 0])
        sns.barplot(y=VALUE_COLUMN, x="valueType", hue="focus", data=bool_df, ax=ax,
                    palette={k: FOCUS_COLORS[k] for k in hue_order},
                    hue_order=hue_order, capsize=0.075, errwidth=VisConfigValues.ERROR_WIDTH, errcolor=VisConfigValues.ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("Percentage Error")
    ax.set_xticklabels(VisConfigValues.VIS_BOOL_ORDER)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VisConfigValues.VIS_VALUES["EB-ylim"])
    ax.set_yticks(VisConfigValues.VIS_VALUES["EB-yTicks"])
    ax.set_yticklabels(VisConfigValues.VIS_VALUES["EB-yLabels"])
    add_xaxis_labels(ax, labels=["Total"], pos="bottom")

    # Bool View - Analysis Task
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 1:])
        sns.barplot(y=VALUE_COLUMN, x="questionType", hue="focus", data=bool_df, ax=ax,
                    palette={k: FOCUS_COLORS[k] for k in hue_order},
                    hue_order=hue_order, order=VisConfigValues.VIS_BOOL_ORDER, capsize=0.075, errwidth=VisConfigValues.ERROR_WIDTH, errcolor=VisConfigValues.ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("")
    ax.set_xticklabels(VisConfigValues.VIS_BOOL_ORDER)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VisConfigValues.VIS_VALUES["EB-ylim"])
    ax.set_yticks(VisConfigValues.VIS_VALUES["EB-yTicks"])
    ax.set_yticklabels([""] * len(VisConfigValues.VIS_VALUES["EB-yLabels"]))
    add_xaxis_labels(ax, labels=VisConfigValues.VIS_BOOL_ORDER, pos="bottom")

    times_df = times_df[times_df.valueDescription == "SolvingDuration"]

    sGs = gs.new_subplotspec(loc=[0, 10], rowspan=1, colspan=9).subgridspec(nrows=1, ncols=9, wspace=0.025)
    # Time - Total
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 0])
        sns.barplot(y=VALUE_COLUMN, x="valueType", hue="focus", data=times_df, ax=ax,
                    palette={k: FOCUS_COLORS[k] for k in hue_order},
                    hue_order=hue_order, capsize=0.075, errwidth=VisConfigValues.ERROR_WIDTH, errcolor=VisConfigValues.ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("Completion Time in s")
    ax.set_xticklabels(VisConfigValues.QUESTION_TYPES)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VisConfigValues.VIS_VALUES["T-ylim"])
    ax.set_yticks(VisConfigValues.VIS_VALUES["T-yTicks"])
    ax.set_yticklabels(VisConfigValues.VIS_VALUES["T-yLabels"])
    add_xaxis_labels(ax, labels=["Total"], pos="bottom")

    # Time - Analysis Task
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 1:])
        sns.barplot(y=VALUE_COLUMN, x="questionType", hue="focus", data=times_df, ax=ax,
                    palette={k: FOCUS_COLORS[k] for k in hue_order},
                    hue_order=hue_order, order=VisConfigValues.QUESTION_TYPES, capsize=0.075, errwidth=VisConfigValues.ERROR_WIDTH, errcolor=VisConfigValues.ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("")
    ax.set_xticklabels(VisConfigValues.QUESTION_TYPES)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VisConfigValues.VIS_VALUES["T-ylim"])
    ax.set_yticks(VisConfigValues.VIS_VALUES["T-yTicks"])
    ax.set_yticklabels([""] * len(VisConfigValues.VIS_VALUES["T-yLabels"]))
    add_xaxis_labels(ax, labels=VisConfigValues.QUESTION_TYPES, pos="bottom")

    add_caption(fig, "B")
    set_legend_from_vis(fig, ax)
    fig.tight_layout()
    for format_str in VIS_FORMAT:
        fig.savefig(f"{IMAGE_OUTPUT_PATH}/focus-questionType.{format_str}")
    plt.close(fig)


def focus_background_questionType(error_df: pd.DataFrame, times_df: pd.DataFrame):
    hue_order = ["BG3 (Single Focus)", "BG3 (Split Focus)", "BG4 (Single Focus)", "BG4 (Split Focus)"]

    df_values = {"fileNr": [], "key": [], "questionType": [], VALUE_COLUMN: [], "valueType": []}
    for _, row in error_df.iterrows():
        df_values["fileNr"].append(row["fileNr"])
        df_values["key"].append(f"{row['background']} ({row['focus']})")
        df_values["questionType"].append(row["questionType"])
        df_values[VALUE_COLUMN].append(row[VALUE_COLUMN])
        df_values["valueType"].append(row["valueType"])
    values_df = pd.DataFrame(df_values)

    numeric_df = values_df[values_df.valueType == "NUMERIC"]
    bool_df = values_df[values_df.valueType == "BOOL"]
    bool_df[VALUE_COLUMN] = bool_df.apply(lambda row: 1 - row[VALUE_COLUMN], axis=1)

    fig = plt.figure(figsize=(FIG_SIZE[0], FIG_SIZE[1] * 1.1), dpi=DPI)
    gs = GridSpec(1, 19, figure=fig, top=TOP_OFFSET, bottom=BOTTOM_OFFSET / 1.1, left=0.02, right=0.99, wspace=1.5)

    sGs = gs.new_subplotspec(loc=[0, 0], rowspan=1, colspan=4).subgridspec(nrows=1, ncols=4, wspace=0.025)
    # Numeric View - Total
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 0])
        sns.barplot(y=VALUE_COLUMN, x="valueType", hue="key", data=numeric_df, ax=ax,
                    palette={k: BG_F_COLORS[k] for k in hue_order},
                    hue_order=hue_order, capsize=0.075, errwidth=VisConfigValues.ERROR_WIDTH, errcolor=VisConfigValues.ERROR_COLOR)

    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("Absolute Errors")
    ax.set_xticklabels(VisConfigValues.VIS_NUMERIC_ORDER)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VisConfigValues.VIS_VALUES["EN-ylim"])
    ax.set_yticks(VisConfigValues.VIS_VALUES["EN-yTicks"])
    ax.set_yticklabels(VisConfigValues.VIS_VALUES["EN-yLabels"])
    add_xaxis_labels(ax, labels=["Total"], pos="bottom")

    # Numeric View - Analysis Task
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 1:])
        sns.barplot(y=VALUE_COLUMN, x="questionType", hue="key", data=numeric_df, ax=ax,
                    palette={k: BG_F_COLORS[k] for k in hue_order},
                    hue_order=hue_order, order=VisConfigValues.VIS_NUMERIC_ORDER, capsize=0.075, errwidth=VisConfigValues.ERROR_WIDTH, errcolor=VisConfigValues.ERROR_COLOR)

    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("")
    ax.set_xticklabels(VisConfigValues.VIS_NUMERIC_ORDER)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VisConfigValues.VIS_VALUES["EN-ylim"])
    ax.set_yticks(VisConfigValues.VIS_VALUES["EN-yTicks"])
    ax.set_yticklabels([""] * len(VisConfigValues.VIS_VALUES["EN-yLabels"]))
    add_xaxis_labels(ax, labels=VisConfigValues.VIS_NUMERIC_ORDER, pos="bottom")

    sGs = gs.new_subplotspec(loc=[0, 4], rowspan=1, colspan=6).subgridspec(nrows=1, ncols=6, wspace=0.025)
    # Bool View - Total
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 0])
        sns.barplot(y=VALUE_COLUMN, x="valueType", hue="key", data=bool_df, ax=ax,
                    palette={k: BG_F_COLORS[k] for k in hue_order},
                    hue_order=hue_order, capsize=0.075, errwidth=VisConfigValues.ERROR_WIDTH, errcolor=VisConfigValues.ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("Percentage Error")
    ax.set_xticklabels(VisConfigValues.VIS_BOOL_ORDER)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VisConfigValues.VIS_VALUES["EB-ylim"])
    ax.set_yticks(VisConfigValues.VIS_VALUES["EB-yTicks"])
    ax.set_yticklabels(VisConfigValues.VIS_VALUES["EB-yLabels"])
    add_xaxis_labels(ax, labels=["Total"], pos="bottom")

    # Bool View - Analysis Task
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 1:])
        sns.barplot(y=VALUE_COLUMN, x="questionType", hue="key", data=bool_df, ax=ax,
                    palette={k: BG_F_COLORS[k] for k in hue_order},
                    hue_order=hue_order, order=VisConfigValues.VIS_BOOL_ORDER, capsize=0.075, errwidth=VisConfigValues.ERROR_WIDTH, errcolor=VisConfigValues.ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("")
    ax.set_xticklabels(VisConfigValues.VIS_BOOL_ORDER)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VisConfigValues.VIS_VALUES["EB-ylim"])
    ax.set_yticks(VisConfigValues.VIS_VALUES["EB-yTicks"])
    ax.set_yticklabels([""] * len(VisConfigValues.VIS_VALUES["EB-yLabels"]))
    add_xaxis_labels(ax, labels=VisConfigValues.VIS_BOOL_ORDER, pos="bottom")

    times_df = times_df[times_df.valueDescription == "SolvingDuration"]
    df_values = {"fileNr": [], "key": [], "questionType": [], VALUE_COLUMN: [], "valueType": []}
    for _, row in times_df.iterrows():
        df_values["fileNr"].append(row["fileNr"])
        df_values["key"].append(f"{row['background']} ({row['focus']})")
        df_values["questionType"].append(row["questionType"])
        df_values[VALUE_COLUMN].append(row[VALUE_COLUMN])
        df_values["valueType"].append(row["valueType"])
    values_df = pd.DataFrame(df_values)

    sGs = gs.new_subplotspec(loc=[0, 10], rowspan=1, colspan=9).subgridspec(nrows=1, ncols=9, wspace=0.025)
    # Time - Total
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 0])
        sns.barplot(y=VALUE_COLUMN, x="valueType", hue="key", data=values_df, ax=ax,
                    palette={k: BG_F_COLORS[k] for k in hue_order},
                    hue_order=hue_order, capsize=0.075, errwidth=VisConfigValues.ERROR_WIDTH, errcolor=VisConfigValues.ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("Completion Time in s")
    ax.set_xticklabels(VisConfigValues.QUESTION_TYPES)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VisConfigValues.VIS_VALUES["T-ylim"])
    ax.set_yticks(VisConfigValues.VIS_VALUES["T-yTicks"])
    ax.set_yticklabels(VisConfigValues.VIS_VALUES["T-yLabels"])
    add_xaxis_labels(ax, labels=["Total"], pos="bottom")

    # Time - Analysis Task
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 1:])
        sns.barplot(y=VALUE_COLUMN, x="questionType", hue="key", data=values_df, ax=ax,
                    palette={k: BG_F_COLORS[k] for k in hue_order},
                    hue_order=hue_order, order=VisConfigValues.QUESTION_TYPES, capsize=0.075, errwidth=VisConfigValues.ERROR_WIDTH, errcolor=VisConfigValues.ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("")
    ax.set_xticklabels(VisConfigValues.QUESTION_TYPES)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VisConfigValues.VIS_VALUES["T-ylim"])
    ax.set_yticks(VisConfigValues.VIS_VALUES["T-yTicks"])
    ax.set_yticklabels([""] * len(VisConfigValues.VIS_VALUES["T-yLabels"]))
    add_xaxis_labels(ax, labels=VisConfigValues.QUESTION_TYPES, pos="bottom")

    add_caption(fig, "C")
    set_legend_from_vis(fig, ax)
    fig.tight_layout()
    for format_str in VIS_FORMAT:
        fig.savefig(f"{IMAGE_OUTPUT_PATH}/focus_background-questionType.{format_str}")
    plt.close(fig)


def questionnaire_general_perception(questionnaire_df: pd.DataFrame):
    questions = [f"generalQuestion{i}{c}BG" for i in range(1, 4) for c in ["First", "Second"]]

    fig = plt.figure(figsize=(FIG_SIZE[0] / 2., FIG_SIZE[1] * (2.5 * 3 / 7.)), dpi=DPI)
    gs = GridSpec(3, 1, figure=fig, top=(1 - (1 - TOP_OFFSET) / (2 * 3 / 7.)), right=0.98, left=0.035, bottom=(BOTTOM_OFFSET - 0.025) / (2 * 3 / 7.),
                  hspace=H_SPACE)

    question_counts = {"background": [], "question": [], 1: [], 2: [], 3: [], 4: [], 5: []}
    for b in questionnaire_df.firstBackground.unique():
        for c in ["First", "Second"]:
            for qi, question in enumerate([v for v in questions if c in v]):
                counts = questionnaire_df[questionnaire_df[f"{c.lower()}Background"] == b][question].value_counts()
                counts = [counts[i] if i in counts else 0 for i in range(1, 6)]

                question_counts["background"].append(b)
                question_counts["question"].append(question.replace(f"{c}BG", ""))
                question_counts[1].append(int(sum(counts[4:5])))
                question_counts[2].append(int(sum(counts[3:5])))
                question_counts[3].append(int(sum(counts[2:5])))
                question_counts[4].append(int(sum(counts[1:5])))
                question_counts[5].append(int(sum(counts[0:5])))

    question_count_df = pd.DataFrame(question_counts, columns=question_counts.keys())
    question_count_df = question_count_df.groupby(by=["background", "question"]).sum().reset_index()

    axes = []
    with sns.axes_style(SNS_AXES_STYLE):
        for i, question in enumerate(list(QUESTION_MAPPING.keys())[0:3]):
            ax = fig.add_subplot(gs[i, 0])

            for j in range(5, 0, -1):
                sns.barplot(y="question", x=j, hue="background",
                            data=question_count_df[question_count_df.question == question],
                            ax=ax, hue_order=["BG3", "BG4"],
                            palette=[PERCEPTION_RATING_COLOR_PALETTE[j - 1] for _ in range(0, 2)])

            ax.set_xlim((0, 16))
            ax.set_xticks(range(0, 17))
            ax.set_xticklabels([""] * 17)
            ax.set_yticklabels([""])
            ax.get_legend().remove()
            ax.set_ylabel("")
            ax.set_xlabel("")
            ax.set_title(QUESTION_MAPPING[question], loc="left")

            add_yaxis_labels_hue(ax, labels=["BG3", "BG4"], y_offset=-0.1)
            axes.append(ax)

        ax.set_xticklabels(range(0, 17))

    set_custom_legend(
        fig=fig,
        ax=ax,
        keys=["agree", "", "neutral", "", "disagree"],
        colors=PERCEPTION_RATING_COLOR_PALETTE,
        title="Five-Point Scale Answers"
    )

    add_caption(fig, "A")
    fig.tight_layout()
    fig.subplots_adjust(top=TOP_OFFSET)
    for format_str in VIS_FORMAT:
        fig.savefig(f"{IMAGE_OUTPUT_PATH}/questionnaire-general-perception-rating.{format_str}")
    plt.close(fig)


def questionnaire_vis_perception(questionnaire_df: pd.DataFrame):
    questions = [f"backgroundPerception{i}{c}BG" for i in range(1, 3) for c in ["First", "Second"]]

    fig = plt.figure(figsize=(FIG_SIZE[0] / 2., FIG_SIZE[1] * (2.5 * 3 / 7.)), dpi=DPI)
    gs = GridSpec(3, 1, figure=fig, top=(1 - (1 - TOP_OFFSET) / (2 * 3 / 7.)), right=0.98, left=0.035, bottom=(BOTTOM_OFFSET - 0.025) / (3 * 2 / 7.),
                  hspace=H_SPACE)

    question_counts = {"background": [], "question": [], 1: [], 2: [], 3: [], 4: [], 5: []}
    for b in questionnaire_df.firstBackground.unique():
        for c in ["First", "Second"]:
            for qi, question in enumerate([v for v in questions if c in v]):
                counts = questionnaire_df[questionnaire_df[f"{c.lower()}Background"] == b][question].value_counts()
                counts = [counts[i] if i in counts else 0 for i in range(1, 6)]

                question_counts["background"].append(b)
                question_counts["question"].append(question.replace(f"{c}BG", ""))
                question_counts[1].append(int(sum(counts[4:5])))
                question_counts[2].append(int(sum(counts[3:5])))
                question_counts[3].append(int(sum(counts[2:5])))
                question_counts[4].append(int(sum(counts[1:5])))
                question_counts[5].append(int(sum(counts[0:5])))

    question_count_df = pd.DataFrame(question_counts, columns=question_counts.keys())
    question_count_df = question_count_df.groupby(by=["background", "question"]).sum().reset_index()

    axes = []
    with sns.axes_style(SNS_AXES_STYLE):
        for i, question in enumerate(list(QUESTION_MAPPING.keys())[3:7]):
            ax = fig.add_subplot(gs[i, 0])

            for j in range(5, 0, -1):
                sns.barplot(y="question", x=j, hue="background",
                            data=question_count_df[question_count_df.question == question],
                            ax=ax, hue_order=["BG3", "BG4"],
                            palette=[PERCEPTION_RATING_COLOR_PALETTE[j - 1] for _ in range(0, 2)])

            ax.set_xlim((0, 16))
            ax.set_xticks(range(0, 17))
            ax.set_xticklabels([""] * 17)
            ax.set_yticklabels([""])
            ax.get_legend().remove()
            ax.set_ylabel("")
            ax.set_xlabel("")
            ax.set_title(QUESTION_MAPPING[question], loc="left")

            add_yaxis_labels_hue(ax, labels=["BG3", "BG4"], y_offset=-0.1)
            axes.append(ax)

        ax.set_xticklabels(range(0, 17))

    set_custom_legend(
        fig=fig,
        ax=ax,
        keys=["greatly", "", "moderate", "", "not at all"],
        colors=PERCEPTION_RATING_COLOR_PALETTE,
        title="Five-Point Scale Answers"
    )

    add_caption(fig, "B")
    fig.tight_layout()
    fig.subplots_adjust(top=TOP_OFFSET)
    for format_str in VIS_FORMAT:
        fig.savefig(f"{IMAGE_OUTPUT_PATH}/questionnaire-vis-perception-rating.{format_str}")
    plt.close(fig)


def questionnaire_focus(questionnaire_df: pd.DataFrame):
    questions = [f"focusQuestion{i}" for i in range(1, 6)]

    fig = plt.figure(figsize=(FIG_SIZE[0] / 2., FIG_SIZE[1] * 5 / 4.), dpi=DPI)
    gs = GridSpec(5, 1, figure=fig, top=TOP_OFFSET, right=0.98, left=0.075, bottom=(BOTTOM_OFFSET - 0.025) / (2 * 5 / 7.),
                  hspace=H_SPACE)

    question_counts = {"question": [], "mean": [], "sd": [], 1: [], 2: [], 3: [], 4: [], 5: []}
    for qi, question in enumerate(questions):
        counts = questionnaire_df[question].value_counts()
        mean = questionnaire_df[question].mean()
        sd = questionnaire_df[question].std()
        counts = [counts[i] if i in counts else 0 for i in range(1, 6)]

        question_counts["question"].append(question)
        question_counts["mean"].append(mean)
        question_counts["sd"].append(sd)
        question_counts[1].append(int(sum(counts[0:1])))
        question_counts[2].append(int(sum(counts[0:2])))
        question_counts[3].append(int(sum(counts[0:3])))
        question_counts[4].append(int(sum(counts[0:4])))
        question_counts[5].append(int(sum(counts[0:5])))

    question_count_df = pd.DataFrame(question_counts, columns=question_counts.keys())

    axes = []
    with sns.axes_style(SNS_AXES_STYLE):
        for i, question in enumerate(FOCUS_QUESTION_MAPPING.keys()):
            ax = fig.add_subplot(gs[i, 0])

            for i in range(5, 0, -1):
                sns.barplot(y="question", x=i,
                            data=question_count_df[question_count_df.question == question],
                            ax=ax,
                            palette=[PERCEPTION_RATING_COLOR_PALETTE[i - 1] for _ in range(0, 3)])

            ax.set_xlim((0, 16))
            ax.set_xticks(range(0, 17))
            ax.set_xticklabels([""] * 17)
            ax.set_yticklabels([""])
            ax.set_ylabel("")
            ax.set_xlabel("")
            ax.set_title(FOCUS_QUESTION_MAPPING[question], loc="left")

            add_yaxis_labels_hue(ax, y_offset=-0.0325, fontsize=24, x_factor=0.0425,
                                 labels=[
                                     f' M  ={question_count_df[(question_count_df.question == question)]["mean"].values[0]:1.2f}\n ' +
                                     f'SD={question_count_df[(question_count_df.question == question)]["sd"].values[0]:1.2f}'
                                 ])

            axes.append(ax)

        ax.set_xticklabels(range(0, 17))

    set_custom_legend(
        fig=fig,
        ax=ax,
        keys=["not at all", "", "moderate", "", "greatly"],
        colors=PERCEPTION_RATING_COLOR_PALETTE,
        title="Five-Point Scale Answers"
    )

    fig.tight_layout()
    fig.subplots_adjust(top=TOP_OFFSET)
    for format_str in VIS_FORMAT:
        fig.savefig(f"{IMAGE_OUTPUT_PATH}/questionnaire-focus-rating.{format_str}")
    plt.close(fig)


if __name__ == '__main__':
    os.makedirs(IMAGE_OUTPUT_PATH, exist_ok=True)

    set_rc_plot_values()

    calculated_values = calc_values()
    time_values = calc_time_values()
    questionnaire_values = get_questionnaire_data()

    calculated_values = calculated_values.replace("AR", "BG3").replace("MK", "BG4")
    calculated_values = calculated_values.replace("Start", "First Quarter").replace("Middle-1", "Second Quarter").replace("Middle-2", "Third Quarter").replace("End", "Fourth Quarter")
    time_values = time_values.replace("AR", "BG3").replace("MK", "BG4")
    time_values = time_values.replace("Start", "First Quarter").replace("Middle-1", "Second Quarter").replace("Middle-2", "Third Quarter").replace("End", "Fourth Quarter")
    calculated_values["focus"] = calculated_values.apply(lambda row: "Split Focus" if row["focus"] is True else "Single Focus", axis=1)
    time_values["focus"] = time_values.apply(lambda row: "Split Focus" if row["focus"] is True else "Single Focus", axis=1)

    calculated_values = calculated_values[~calculated_values.questionType.isin(["Filter Errors", "Sort Errors", "Determine Range", "-"])]
    calculated_values = calculated_values.replace("Range Values", "Range").replace("Compute Derived Value", "Compute").replace("Retrieve Value", "Retrieve")
    time_values = time_values.replace("Determine Range", "Range").replace("Compute Derived Value", "Compute").replace("Retrieve Value", "Retrieve")

    background_questionType(calculated_values, time_values)
    focus_questionType(calculated_values, time_values)
    focus_background_questionType(calculated_values, time_values)

    questionnaire_general_perception(questionnaire_values)
    questionnaire_vis_perception(questionnaire_values)
    questionnaire_focus(questionnaire_values)
