import math
from typing import List

import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.patches import Patch

from config import TOP_BORDER_OFFSET, BOTTOM_BORDER_OFFSET, RC_FONTS, CAPTION_COLOR, FIG_SIZE, DPI


def find_indices(data: List[List[str]], value: object, column_index: int) -> List[int]:
    output = []
    for i, r in enumerate(data):
        if value in r[column_index]:
            output.append(i)
    return output


def data_string_converter(data_string: str) -> object:
    if type(data_string) is not str:
        if math.isnan(data_string):
            return None
        else:
            return data_string

    if data_string == "":
        return ""
    elif "[" in data_string and "|" not in data_string:
        values = data_string[1:-1].split(',')
        return [int(v) for v in values]
    elif "(" in data_string and "|" not in data_string:
        values = data_string[1:-1].split(',')
        return tuple(int(v) for v in values)
    elif "True" in data_string:
        return True
    elif "False" in data_string:
        return False
    else:
        try:
            if "." in data_string:
                value = float(data_string)
            else:
                value = int(data_string)
            return value
        except:
            return data_string


# Utils for the visualization creation
def set_rc_plot_values():
    plt.rc('font', size=RC_FONTS["font"])  # controls default text sizes
    plt.rc('axes', titlesize=RC_FONTS["axes"]["title"])  # fontsize of the axes title
    plt.rc('axes', labelsize=RC_FONTS["axes"]["label"], labelweight="bold")  # fontsize of the x and y labels
    plt.rc('xtick', labelsize=RC_FONTS["xtick"])  # fontsize of the tick labels
    plt.rc('ytick', labelsize=RC_FONTS["ytick"])  # fontsize of the tick labels
    plt.rc('legend', fontsize=RC_FONTS["legend"])  # legend fontsize
    plt.rc('figure', titlesize=RC_FONTS["figure"])  # fontsize of the figure title
    # sns.set_context(SNS_CONTEXT)

def add_label_to_bars(input_df: pd.DataFrame, ax: plt.Axes, x_order: List,
                      hue_key: str, x_order_key: str, value_key: str, value_type=float,
                      rotation: float = 0, factor: float = 1):
    input_df[value_key] = input_df[value_key].astype("float64")
    df = input_df.groupby(by=[x_order_key, hue_key]).mean().reset_index()
    for bars in ax.containers:
        for bar, key in zip(bars, x_order):
            try:
                if value_type is int:
                    label = f"{list(df[(df[hue_key] == bars._label) & (df[x_order_key] == key)][value_key])[0] * factor}".split('.')[0]
                else:
                    label = f"{list(df[(df[hue_key] == bars._label) & (df[x_order_key] == key)][value_key])[0] * factor:1.2f}"
            except:
                label = 0

            ax.text(
                x=bar.get_x() + bar.get_width() / 2,
                # y=ax.viewLim.y1 * 0.025,
                y=ax.viewLim.y1 * -0.02,
                s=label,
                ha="center",
                va="top",
                color="#000000",
                rotation=rotation
            )

def set_legend_from_vis(fig: plt.Figure, ax: plt.Axes, label_on_fig: bool = True):
    obj = fig if label_on_fig else ax
    handles, labels = ax.get_legend_handles_labels()
    obj.legend(
        handles,
        labels,
        loc="upper center",
        bbox_to_anchor=(0.5, 1),
        ncol=5,
        framealpha=0.95,
        edgecolor="#000000"
    )


def set_custom_legend(fig: plt.Figure, ax: plt.Axes, keys: List[str], colors: List[str],
                      title: str = None, label_on_fig: bool = True):

    obj = fig if label_on_fig else ax
    custom_lines = [Patch(facecolor="#ffffff", edgecolor="#ffffff")] + [Patch(facecolor=c, edgecolor=c) for c in colors]
    leg = obj.legend(
        custom_lines,
        [title + ":"] + keys,
        # markerfirst=False,
        loc="upper center",
        bbox_to_anchor=(0.5, 1),
        ncol=6,
        framealpha=0.95,
        edgecolor="#000000"
    )
    for vpack in leg._legend_handle_box.get_children()[:1]:
        for hpack in vpack.get_children():
            hpack.get_children()[0].set_width(0)


def add_xaxis_labels(ax: plt.Axes, labels: List, pos: str = "top"):
    if pos == "top":
        y = ax.viewLim.y1 * TOP_BORDER_OFFSET
    elif pos == "bottom":
        y = ax.viewLim.y0 - ax.viewLim.y1 * BOTTOM_BORDER_OFFSET
    else:
        return

    for i, label in enumerate(labels):
        pos0 = ax.containers[0][i].get_x()
        pos1 = ax.containers[-1][i].get_x()
        x = pos0 + (pos1 - pos0) / 2
        width = ax.containers[0][i].get_width()

        ax.text(
            x=x + width / 2.,
            y=y,
            s=label,
            ha="center",
            va="bottom",
            color="#000000"
        )


def add_yaxis_labels_hue(ax: plt.Axes, labels: List, fontsize: int = 26, y_offset: float = 0.0, x_factor: float = 0.0175):
    for i, label in enumerate(labels):
        bar = ax.containers[i][0]

        ax.text(
            x=ax.viewLim.x0 - ax.viewLim.x1 * x_factor,
            y=bar.get_y() + bar.get_height() + y_offset,
            s=label,
            ha="center",
            va="bottom",
            color="#000000",
            fontsize=fontsize,
        )


def add_caption(fig: plt.Figure, caption: str):
    size = fig.get_size_inches()
    fig.patches.extend([plt.Rectangle(
        xy=(0, 1),
        width=0.75 / size[0],
        height=-0.75 / size[1],
        fill=True,
        color=CAPTION_COLOR,
        transform=fig.transFigure
    )])
    fig.text(
        x=(0.75 / size[0]) / 2 - 0.0005,
        y=1 - (0.75 / size[1]) / 2 - 0.005,  # 0.005 for Study 2 questionnaires
        s=caption,
        fontsize=61,
        ha="center",
        va="center",
        weight="bold",
        fontname="calibri",
        color="white",
        transform=fig.transFigure
    )
